@@ -1,7 +1,6 @@
 # Imports
 
 import concurrent.futures
-import functools
 import heapq
 import re
 import typing
@@ -15,6 +14,7 @@ import pandas as pd
 import sklearn.feature_extraction.text as sklearn_text
 import sklearn.neighbors
 from thefuzz import fuzz
+import tqdm
 
 EditDistanceScorers = Literal[
     "damerau_levenshtein",
@@ -51,6 +51,201 @@ DEFAULT_SCORERS = {
     "tfidf": .8
 }
 
+# Functions used in both classes
+def _get_score_from_distance(
+    lookup_str: str,
+    original_str: str,
+    original_str_len: int,
+    scorer: Callable
+) -> tuple[float, str, str]:
+    dist = scorer(original_str, lookup_str)
+
+    # For methods that return a distance instead of a percentage, divide the returned distance
+    # by the length of the longest string to obtain a relative percentage. Otherwise, the
+    # distance when comparing long string to short strings will be biased against long srings.
+    max_len = max(original_str_len, len(lookup_str))
+    pct = (max_len - dist) / max_len
+
+    return (pct, original_str, lookup_str)
+
+def _get_score(lookup_str: str, original_str: str, scorer: Callable) -> tuple[float, str, str]:
+    pct = scorer(original_str, lookup_str)
+
+    return (pct, original_str, lookup_str)
+
+class Comparer:
+
+    def __init__(self, original: pd.Series, lookup: pd.Series):
+        self._original = original
+        self._lookup = lookup
+
+        self._validate()
+
+    def _validate(self):
+        og_len = len(self._original.index)
+        lu_len = len(self._lookup.index)
+
+        if og_len == 0 or lu_len == 0:
+            raise ValueError("Both series must be non-empty.")
+
+        if og_len != lu_len:
+            raise ValueError("Column lengths do not match.")
+
+        for col in (self._original, self._lookup):
+            if not pd.api.types.is_string_dtype(col):
+                raise TypeError(f"Column {col.name} is not string type.")
+
+    def _get_score_from_distance(self, o_str: str, lu_str: str, scorer: Callable) -> float:
+        if o_str == "" or lu_str == "":
+            return (0, o_str, lu_str)
+
+        o_str_len = len(o_str)
+
+        return _get_score_from_distance(o_str, lu_str, o_str_len, scorer)
+
+    def _get_score(self, o_str: str, lu_str: str, scorer: Callable) -> float:
+        return _get_score(o_str, lu_str, scorer)
+
+    def _apply_score(
+        self,
+        match_func: Callable,
+        scorer: Callable,
+        already_ratio: bool=False
+    ) -> pd.Series:
+        original = self._original.to_list() # faster to iterate over list
+        lookup = self._lookup.to_list() 
+        
+        score_lst = []
+        with tqdm.tqdm(total=len(self._original.index)) as pbar:
+            with concurrent.futures.ProcessPoolExecutor() as pool:
+                futures = [
+                    pool.submit(match_func, o_str, lu_str, scorer) for o_str, lu_str \
+                        in zip(original, lookup)
+                ]
+
+                for future in concurrent.futures.as_completed(futures):
+                    score, _, _ = future.result()
+                    score_lst.append(score)
+
+                    pbar.update(1)
+
+        res = pd.Series(score_lst)
+        
+        if already_ratio:
+            return res
+
+        res = res * 100
+
+        return res
+
+    def damerau_levenshtein(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score_from_distance,
+            scorer=jellyfish.damerau_levenshtein_distance
+        )
+
+    def levenshtein(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score_from_distance,
+            scorer=Levenshtein.distance
+        )
+
+    def hamming(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score_from_distance,
+            scorer=jellyfish.hamming_distance
+        )
+
+    def jaro_winkler(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score,
+            scorer=jarowinkler.jarowinkler_similarity
+        )
+
+    def jaro(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score,
+            scorer=jarowinkler.jaro_similarity
+        )
+
+    def ratio(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score,
+            scorer=fuzz.ratio,
+            already_ratio=True
+        )
+
+    def partial_ratio(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score,
+            scorer=fuzz.partial_ratio,
+            already_ratio=True
+        )
+
+    def token_sort_ratio(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score,
+            scorer=fuzz.token_sort_ratio,
+            already_ratio=True
+        )
+
+    def token_set_ratio(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score,
+            scorer=fuzz.token_set_ratio,
+            already_ratio=True
+        )
+
+    def partial_token_set_ratio(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score,
+            scorer=fuzz.partial_token_set_ratio,
+            already_ratio=True
+        )
+
+    def partial_token_sort_ratio(self) -> pd.Series:
+        return self._apply_score(
+            match_func=self._get_score,
+            scorer=fuzz.partial_token_sort_ratio,
+            already_ratio=True
+        )
+
+    def _dispatcher(self, func_name: str) -> Callable:
+        func_mapper = {
+            "damerau_levenshtein": self.damerau_levenshtein,
+            "hamming": self.hamming,
+            "jaro": self.jaro,
+            "jaro_winkler": self.jaro_winkler,
+            "levenshtein": self.levenshtein,
+            "partial_ratio": self.partial_ratio,
+            "partial_token_set_ratio": self.partial_token_set_ratio,
+            "partial_token_sort_ratio": self.partial_token_sort_ratio,
+            "ratio": self.ratio,
+            "token_set_ratio": self.token_set_ratio,
+            "token_sort_ratio": self.token_sort_ratio,
+        }
+
+        return func_mapper[func_name]
+
+    def compare(
+        self, 
+        scorers: EditDistanceScorers, 
+        weights: list[int | float]
+    ) -> pd.Series:
+        sum_w = sum(weights)
+        weights = [w / sum_w for w in weights] # normalize to 1
+        scorer_dict = {k: v for k, v in zip(scorers, weights)}
+        
+        scores = dict()
+        for func_nm, weight in scorer_dict.items():
+            func = self._dispatcher(func_nm)
+            scores[f"{func_nm}_score"] = func() * weight
+
+        df = pd.DataFrame(scores)
+        df["final_score"] = df[list(scores.keys())].sum(axis=1, skipna=True)
+
+        return df["final_score"]
+
 class Matcher:
 
     def __init__(self, original: pd.Series, lookup: pd.Series) -> None:
@@ -186,135 +381,111 @@ class Matcher:
     def match_rating_codex(self) -> pd.DataFrame:
         return self._phonetic_match(jellyfish.match_rating_codex, "mr_codex")
 
-    @staticmethod
-    def _get_score_from_distance(
-        lookup_str: str,
-        original_str: str,
-        original_str_len: int,
-        scorer: Callable
-    ) -> float:
-        dist = scorer(original_str, lookup_str)
+    def _get_matches_distance(
+        self, 
+        o_str: str, 
+        lookup: list[str], 
+        scorer: Callable, 
+        k_matches: int
+    ) -> list[tuple]:
+        o_str_len = len(o_str)
+        matches = [_get_score_from_distance(lu_str, o_str, o_str_len, scorer) for lu_str in lookup]
+        matches = heapq.nlargest(k_matches, matches)
 
-        # For methods that return a distance instead of a percentage, divide the returned distance
-        # by the length of the longest string to obtain a relative percentage. Otherwise, the
-        # distance when comparing long string to short strings will be biased against long srings.
-        max_len = max(original_str_len, len(lookup_str))
-        pct = (max_len - dist) / max_len
-
-        return (pct, lookup_str)
-
-    @staticmethod
-    def _get_score(lookup_str: str, original_str: str, scorer: Callable) -> float:
-        pct = scorer(original_str, lookup_str)
-
-        return (pct, lookup_str)
+        return matches
 
     def _get_matches_pct(
         self, 
+        o_str: str, 
+        lookup: list[str], 
         scorer: Callable, 
+        k_matches: int
+    ) -> list[tuple]:
+        matches = [_get_score(lu_str, o_str, scorer) for lu_str in lookup]
+        matches = heapq.nlargest(k_matches, matches)
+
+        return matches
+
+    def _get_all_matches(
+        self,
+        match_func: Callable,
+        scorer: Callable,
         k_matches: int, 
         threshold: int,
         already_ratio: bool=False
     ) -> pd.DataFrame:
-        original = self._original
-        lookup = self._lookup
+        original = self._original.to_list() # faster to iterate over list
+        lookup = self._lookup.to_list() 
         
-        # Maybe premature optimization, but let's store all the things we will need in variables
-        # up front, instead of paying the cost of, e.g. lookup in the loop
-        og_colname = original.name
-        lu_colname = lookup.name
+        og_colname = self._original.name
+        lu_colname = self._lookup.name
 
         merged = {og_colname: [], lu_colname: [], "score": []}
-        for o_str in original:
-            with concurrent.futures.ThreadPoolExecutor() as pool:
-                func = functools.partial(
-                    self._get_score,
-                    original_str=o_str,
-                    scorer=scorer
-                )
-                matches = pool.map(func, lookup)
+        with tqdm.tqdm(total=len(self._original.index)) as pbar:
+            with concurrent.futures.ProcessPoolExecutor() as pool:
+                futures = [
+                    pool.submit(match_func, o_str, lookup, scorer, k_matches) \
+                        for o_str in original
+                ]
+                for future in concurrent.futures.as_completed(futures):
+                    matches = future.result()
+                    for score, o_str, lu_str in matches:
+                        merged[og_colname].append(o_str)
+                        merged[lu_colname].append(lu_str)
+                        merged["score"].append(score)
 
-            matches = heapq.nlargest(k_matches, matches)
-            for score, lu_name in matches:
-                merged[og_colname].append(o_str)
-                merged[lu_colname].append(lu_name)
-                merged["score"].append(score)
+                    pbar.update(1)
 
         merged = pd.DataFrame(merged)
-
+        
         if already_ratio:
             merged["score"] = merged["score"] / 100
 
         return self._clean_and_filter(merged, threshold)
 
-    def _get_matches_distance(self, scorer: Callable, k_matches: int, threshold: int) -> pd.DataFrame:
-        original = self._original
-        lookup = self._lookup
-        
-        # Maybe premature optimization, but let's store all the things we will need in variables
-        # up front, instead of paying the cost of, e.g. lookup in the loop
-        og_colname = original.name
-        lu_colname = lookup.name
-
-        merged = {og_colname: [], lu_colname: [], "score": []}
-        for o_str in original:
-            o_str_len = len(o_str)
-            with concurrent.futures.ThreadPoolExecutor() as pool:
-                func = functools.partial(
-                    self._get_score_from_distance,
-                    original_str=o_str,
-                    original_str_len=o_str_len,
-                    scorer=scorer
-                )
-                matches = pool.map(func, lookup)
-
-            matches = heapq.nlargest(k_matches, matches)
-            for score, lu_name in matches:
-                merged[og_colname].append(o_str)
-                merged[lu_colname].append(lu_name)
-                merged["score"].append(score)
-
-        merged = pd.DataFrame(merged)
-
-        return self._clean_and_filter(merged, threshold)
-
     def damerau_levenshtein(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_distance(
+        return self._get_all_matches(
+            match_func=self._get_matches_distance,
             scorer=jellyfish.damerau_levenshtein_distance,
             k_matches=k_matches, 
             threshold=threshold
         )
 
     def levenshtein(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_distance(
+        return self._get_all_matches(
+            match_func=self._get_matches_distance,
             scorer=Levenshtein.distance,
             k_matches=k_matches, 
             threshold=threshold
         )
 
     def hamming(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_distance,
             scorer=jellyfish.hamming_distance,
             k_matches=k_matches,
             threshold=threshold
         )
 
     def jaro_winkler(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_pct,
             scorer=jarowinkler.jarowinkler_similarity,
             k_matches=k_matches,
             threshold=threshold
         )
 
     def jaro(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_pct,
             scorer=jarowinkler.jaro_similarity,
             k_matches=k_matches,
             threshold=threshold
         )
 
     def ratio(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_pct,
             scorer=fuzz.ratio,
             k_matches=k_matches,
             threshold=threshold,
@@ -322,7 +493,8 @@ class Matcher:
         )
 
     def partial_ratio(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_pct,
             scorer=fuzz.partial_ratio,
             k_matches=k_matches,
             threshold=threshold,
@@ -330,7 +502,8 @@ class Matcher:
         )
 
     def token_sort_ratio(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_pct,
             scorer=fuzz.token_sort_ratio,
             k_matches=k_matches,
             threshold=threshold,
@@ -338,7 +511,8 @@ class Matcher:
         )
 
     def token_set_ratio(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_pct,
             scorer=fuzz.token_set_ratio,
             k_matches=k_matches,
             threshold=threshold,
@@ -346,7 +520,8 @@ class Matcher:
         )
 
     def partial_token_set_ratio(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_pct,
             scorer=fuzz.partial_token_set_ratio,
             k_matches=k_matches,
             threshold=threshold,
@@ -354,7 +529,8 @@ class Matcher:
         )
 
     def partial_token_sort_ratio(self, k_matches: int=5, threshold: int=80) -> pd.DataFrame:
-        return self._get_matches_pct(
+        return self._get_all_matches(
+            match_func=self._get_matches_pct,
             scorer=fuzz.partial_token_sort_ratio,
             k_matches=k_matches,
             threshold=threshold,
